# coding: utf-8
"""
This script belongs to the medenv package
Copyright (C) 2022 Jeremy Fix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# Standard imports
import pathlib
import urllib
import shutil
import gzip
import logging

# External imports
import tqdm

_BASEDIR = pathlib.Path.home() / ".medenv"

# my_hook comes for the tqdm examples


def my_hook(t):
    """Wraps tqdm instance.
    Don't forget to close() or __exit__()
    the tqdm instance once you're done with it (easiest using `with` syntax).
    Example
    -------
    >>> with tqdm(...) as t:
    ...     reporthook = my_hook(t)
    ...     urllib.urlretrieve(..., reporthook=reporthook)
    """
    last_b = [0]

    def update_to(b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] or -1,
            remains unchanged.
        """
        if tsize not in (None, -1):
            t.total = tsize
        displayed = t.update((b - last_b[0]) * bsize)
        last_b[0] = b
        return displayed

    return update_to


def download_url(url, filename=None):
    logging.debug(f"Downloading from {url}")
    if filename is not None and filename.exists():
        return filename
    with tqdm.tqdm(unit="B", unit_scale=True, unit_divisor=1024, miniters=1) as t:
        local_filename, headers = urllib.request.urlretrieve(
            url, filename, reporthook=my_hook(t)
        )
    return local_filename


def _extract_gz(filepath: pathlib.Path, target: pathlib.Path):
    with gzip.open(filepath, "rb") as f_in:
        with open(target, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)


_ARCHIVE_EXTRACTORS = {"gzip": _extract_gz, "raw": lambda: None}


def download_and_extract(url, archivetype, name):
    if not _BASEDIR.exists():
        _BASEDIR.mkdir()
    f_out = _BASEDIR / name
    if f_out.exists():
        logging.debug(f"Data already extracted to {f_out}")
        return f_out

    local_filename = download_url(url)
    logging.debug(f"Data temporarily saved to {local_filename}")
    _ARCHIVE_EXTRACTORS[archivetype](local_filename, f_out)
    logging.debug(f"Data extracted to {f_out}")
    return f_out



# coding: utf-8
"""
This script belongs to the medenv package
Copyright (C) 2022 Jeremy Fix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# The data are provided by the NOAA
# see : https://www.ncei.noaa.gov/products/world-ocean-atlas

# Standard imports
import logging
from datetime import datetime

# External imports
import pandas as pd

# Local imports
from medenv import utils


# Available measures from WOA 2018
# Temperature (t), Salinity (s), Oxygen (o), Nitrate(n), Phosphate(p), Silicate(i)
# For mixed layer depth, Density, Conductivity, Mixed Layer Depth, Dissolved Oxygen, Percent Oxygen Saturation, Apprent Oxygen utilization, see : https://www.ncei.noaa.gov/access/world-ocean-atlas-2018/
_WOA_BASE_URL = "https://www.ncei.noaa.gov/data/oceans/woa/WOA18/DATA/"
# temperature/csv/8594/1.00/woa18_8594_t01mn01.csv.gz

_WOA_LAND_SEA_URL = (
    "https://www.ncei.noaa.gov/data/oceans/woa/WOA18/MASKS/landsea_04.msk"
)

_AVAILABLE_RESOLUTIONS = [5.0, 1.0, 0.25]
_DEFAULT_RESOLUTION = 1.0
_AVAILABLE_MEASURES = {
    "temperature": "t",
    "salinity": "s",
    # "oxygen": "o",  # Not yet released for WOA2018 [2022]
    # "nitrate": "t",
    # "phosphate": "p",
    # "silicate": "i",
}


def get_period(year):
    if year < 1955 or year > 2017:
        raise ValueError(
            f"WOA2018 does not provide measures for year {year}. The available years are in 1955-2017"
        )
    elif year <= 1964:
        return "5564"
    elif year <= 1974:
        return "6574"
    elif year <= 1984:
        return "7584"
    elif year <= 1994:
        return "8594"
    elif year <= 2004:
        return "95A4"
    elif year <= 2017:
        return "A5B7"


def get_grid_resolution_id(resolution):
    if resolution == 0.25:
        return "04"
    elif resolution == 1.0:
        return "01"
    elif resolution == 5.0:
        return "5d"
    else:
        raise ValueError(
            f"Resolution {resolution} not supported by WOA. Possible resolutions are 0.25, 1.0 or 5.0"
        )


def get_statistics_id(statistics):
    dstatistics = {
        "Objectively analyzed climatology": "an",
        "Statistical mean": "mn",  #
        "Number of observations": "dd",
        "Seasonal climatology": "ma",  # or monthly climatology
        "Standard deviation": "sd",
        "Standard error": "se",  # Standard deviation from statistical mean
        "Statistical mean moa": "oa",  # Statistical mean minus objectively analyzed climatology
        "Number of mean values": "gp",  # Number of mean values within radius of influence
    }
    return dstatistics[statistics]


def build_url(date, what, resolution):
    # The datafile naming convention follows :
    # woa18_[DECA]_[v][tp][ft][gr].[form_end]
    period = get_period(date.year)
    month = date.month
    whatkey = _AVAILABLE_MEASURES[what]
    statistics = get_statistics_id("Objectively analyzed climatology")
    grid_resolution = get_grid_resolution_id(resolution)
    assert resolution in _AVAILABLE_RESOLUTIONS
    return f"{_WOA_BASE_URL}/{what}/csv/{period}/{resolution:.02f}/woa18_{period}_{whatkey}{month:02d}{statistics}{grid_resolution}.csv.gz"


def read_csv(filepath):
    with open(filepath, "r") as f:
        f.readline()  # skip the header
        # The header containings the depths are
        # given like :
        #   #COMMA SEPARATED LATITUDE, LONGITUDE, AND VALUES AT DEPTHS (M):0,5,10, ...
        header_depth = f.readline()
        depths = list(map(float, header_depth.split(":")[1].split(",")))
        data = pd.read_csv(
            f, encoding="iso-8859-1", names=["latitude", "longitude"] + depths
        )
        return data


def fetch_values(date, what, resolution=_DEFAULT_RESOLUTION):
    """
    (longitude, latitude) :(degrees east, degrees north)
    date: datetime
    what: str in ['SST', 'Tmax-min', 'Salinity', 'MLD', ...]
    resolution: 5.00, 1.00 or 0.25
    """

    url = build_url(date, what, resolution)
    filename = f"r{resolution}-" + url.split("/")[-1][:-3]  # file.csv
    filepath = utils.download_and_extract(url, "gzip", filename)
    data = read_csv(filepath)
    return data


def get_value(long_lat, depth, what, resolution=_DEFAULT_RESOLUTION):
    data = fetch_values(datetime(year=2016, month=1, day=1), what, resolution)
    long0, lat0 = long_lat
    idx = (
        abs(data["longitude"] - long0) ** 2 + abs(data["latitude"] - lat0) ** 2
    ).idxmin()
    depth_idx = abs(data.columns[2:] - depth).argmin()
    # TODO: warn if the depth_idx is the maximal depth ?? maybe far from the request ?
    return data.loc[idx].iloc[depth_idx + 2]


def fetch_landsea():
    filename = "land_sea.msk"
    filename = utils._BASEDIR / filename
    filepath = utils.download_url(_WOA_LAND_SEA_URL, filename)
    data = pd.read_csv(
        filepath,
        encoding="iso-8859-1",
        names=["latitude", "longitude", "BottomStdLevel"],
        skiprows=2,
    )
    return data


def is_land(long_lat):
    data = fetch_landsea()
    long0, lat0 = long_lat
    idx = (
        abs(data["longitude"] - long0) ** 2 + abs(data["latitude"] - lat0) ** 2
    ).idxmin()
    return data.loc[idx].iloc[2] == 1  # 1 for land / 0 for sea



# coding: utf-8
"""
This script belongs to the medenv package
Copyright (C) 2022 Jeremy Fix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Standard imports
import logging

# External imports
import netCDF4
import numpy as np

# Local imports
from medenv import utils

# doi:10.7289/V5C8276M
BEDROCK_GRID_URL = "https://www.ngdc.noaa.gov/mgg/global/relief/ETOPO1/data/bedrock/grid_registered/netcdf/ETOPO1_Bed_g_gmt4.grd.gz"

# The netCDF dataset id
#     Conventions: COARDS/CF-1.0
#    title: ETOPO1_Bed_g_gmt4.grd
#    GMT_version: 4.4.0
#    node_offset: 0
#    dimensions(sizes): x(21601), y(10801)
#    variables(dimensions): float64 x(x), float64 y(y), int32 z(y, x)
# with
# x : longitude,
#    long_name: Longitude
#    actual_range: [-180.  180.]
#    units: degrees_east
# y : latitude
#    long_name: Latitude
#    actual_range: [-90.  90.]
#    units: degrees_north
#
dataset = None
longvals = None
latvals = None
depthvals = None
derivate_dephvals = None


def fetch_values():
    global dataset, longvals, latvals, depthvals, derivative_depthvals
    if dataset is None:
        # Download the data archive and extract it
        logging.debug("Downloading the etopo1 data doi:10.7289/V5C8276M")
        filepath = utils.download_and_extract(BEDROCK_GRID_URL, "gzip", "eotop1.grd")
        # Load the CDF file
        dataset = netCDF4.Dataset(filepath)
        longvals = dataset.variables["x"][:]
        latvals = dataset.variables["y"][:]
        depthvals = dataset.variables["z"][:]
        # Compute the detta in longitude/latitude, assuming
        # constant sampling in each direction
        dlongitude = longvals[1] - longvals[0]
        dlatitude = latvals[1] - latvals[0]
        # The computation of the gradient takes some times
        # Maybe we could reduce this computational time by
        # restricting the values fetched
        # e.g. restricting to the mediterranean sea
        derivatives_depthvals = np.gradient(depthvals, dlongitude, dlatitude)
        derivative_depthvals = np.sqrt(
            derivatives_depthvals[0] ** 2 + derivatives_depthvals[1] ** 2
        )
        logging.debug(f"ETOPO1 netCDF dataset informations : \n {dataset}")


def get_value(longitude, latitude):
    """
    Return the depth from the closest longitude, latitude in etopo1

    Args:
        longitude : degrees east
        latitude : degrees north
    """
    fetch_values()
    assert (longvals.min() <= longitude <= longvals.max()) and (
        latvals.min() <= latitude <= latvals.max()
    )
    longidx = np.fabs(longvals - longitude).argmin()
    latidx = np.fabs(latvals - latitude).argmin()
    depth = depthvals[latidx, longidx]
    return depth, {"longitude": longvals[longidx], "latitude": latvals[latidx]}


def get_dvalue(longitude, latitude):
    """
    Return the gradient of the depth from the closest longitude, latitude in etopo1

    Args:
        longitude : degrees east
        latitude : degrees north
    """
    fetch_values()
    assert (longvals.min() <= longitude <= longvals.max()) and (
        latvals.min() <= latitude <= latvals.max()
    )
    longidx = np.fabs(longvals - longitude).argmin()
    latidx = np.fabs(latvals - latitude).argmin()
    d_depth = derivative_depthvals[latidx, longidx]
    return d_depth, {"longitude": longvals[longidx], "latitude": latvals[latidx]}


def is_land(long_lat):
    # This is_land is pretty long,
    # prefer using the woa.is_land function
    long0, lat0 = long_lat
    return get_value(long0, lat0) >= 0



# coding: utf-8
"""
This script belongs to the medenv package
Copyright (C) 2022 Jeremy Fix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# Standard imports
import logging
from datetime import datetime
import functools
from typing import Union

logging.getLogger("medenv").addHandler(logging.NullHandler())


# Local imports
from medenv import etopo
from medenv import woa
from medenv import cmems


class Fetcher:
    _available_features = [
        "bathymetry",
        "sea-surface-temperature",
        "sea-surface-salinity",
    ] + list(cmems.CMEMS._feature_params.keys())

    def __init__(self, features, reduction=None):
        self.features = features
        self.getters = {}
        self.cmems_getter = None
        print(len(Fetcher._available_features))
        for f in features:
            if f not in Fetcher._available_features:
                raise ValueError(
                    f"The feature {f} is not in the available features {Fetcher._available_features}"
                )

            if f in list(cmems.CMEMS._feature_params.keys()):
                if self.cmems_getter is None:
                    self.cmems_getter = cmems.CMEMS()
                self.getters[f] = functools.partial(
                    self.cmems_getter.get_value, what=f, reduction=reduction
                )
            elif f == "bathymetry":
                self.getters[f] = lambda date, long_lat, depth: etopo.get_value(
                    long_lat[0], long_lat[1]
                )
            elif f == "sea-surface-temperature":
                if self.cmems_getter is None:
                    self.cmems_getter = cmems.CMEMS()
                self.getters[
                    f
                ] = lambda date, long_lat, depth: self.cmems_getter.get_value(
                    date=date,
                    long_lat=long_lat,
                    depth=0,
                    what="temperature",
                    reduction=reduction,
                )
            elif f == "sea-surface-salinity":
                if self.cmems_getter is None:
                    self.cmems_getter = cmems.CMEMS()
                self.getters[
                    f
                ] = lambda date, long_lat, depth: self.cmems_getter.get_value(
                    date=date,
                    long_lat=long_lat,
                    depth=0,
                    what="salinity",
                    reduction=reduction,
                )

    def get_values(
        self,
        date: Union[datetime, tuple[datetime, datetime]],
        long_lat: tuple[float, float],
        depth: Union[float, tuple[float, float]],
        reduction=None,
    ):
        values, infos = {}, {}
        for f in self.features:
            logging.info(f"Fetching {f} ")
            values[f], infos[f] = self.getters[f](date, long_lat, depth)
        return values, infos



# coding: utf-8
"""
This script belongs to the medenv package
Copyright (C) 2022 Jeremy Fix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Standard imports
import logging
import datetime
import os
from typing import Union

# External modules
import getpass
import xarray as xr

from pydap.client import open_url
from pydap.cas.get_cookies import setup_session


class CMEMS(object):
    # Datasets used for accessing the measurements
    # med-cmcc
    # https://resources.marine.copernicus.eu/product-detail/MEDSEA_MULTIYEAR_PHY_006_004/INFORMATION
    # med-ogs :
    # https://resources.marine.copernicus.eu/product-detail/MEDSEA_MULTIYEAR_BGC_006_008/INFORMATION
    _feature_params = {
        "temperature": {
            "dataset": "med-cmcc-tem-rean-d",
            "prefix": "my",
            "field": "thetao",
            "slice_mode": "lon-lat",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1987", "%d-%m-%Y"),
        },
        "salinity": {
            "dataset": "med-cmcc-sal-rean-d",
            "prefix": "my",
            "field": "so",
            "slice_mode": "lon-lat",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1987", "%d-%m-%Y"),
        },
        "eastward-water-velocity": {
            "dataset": "med-cmcc-cur-rean-d",
            "prefix": "my",
            "field": "uo",
            "slice_mode": "lon-lat",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1987", "%d-%m-%Y"),
        },
        "northward-water-velocity": {
            "dataset": "med-cmcc-cur-rean-d",
            "prefix": "my",
            "field": "vo",
            "slice_mode": "lon-lat",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1987", "%d-%m-%Y"),
        },
        "mixed-layer-thickness": {
            "dataset": "med-cmcc-mld-rean-d",
            "prefix": "my",
            "field": "mlotst",
            "slice_mode": "lon-lat",
            "has_depth": False,
            "date_limit": datetime.datetime.strptime("01-01-1987", "%d-%m-%Y"),
        },
        "sea-surface-above-geoid": {
            "dataset": "med-cmcc-ssh-rean-d",
            "prefix": "my",
            "field": "zos",
            "slice_mode": "lon-lat",
            "has_depth": False,
            "date_limit": datetime.datetime.strptime("01-01-1987", "%d-%m-%Y"),
        },
        "phytoplankton-carbon-biomass": {
            "dataset": "med-ogs-pft-rean-d",
            "prefix": "my",
            "field": "phyc",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "chlorophyl-a": {
            "dataset": "med-ogs-pft-rean-d",
            "prefix": "my",
            "field": "chl",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "nitrate": {
            "dataset": "med-ogs-nut-rean-d",
            "prefix": "my",
            "field": "no3",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "phosphate": {
            "dataset": "med-ogs-nut-rean-d",
            "prefix": "my",
            "field": "po4",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "ammonium": {
            "dataset": "med-ogs-nut-rean-d",
            "prefix": "my",
            "field": "nh4",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "net-primary-production": {
            "dataset": "med-ogs-bio-rean-d",
            "prefix": "my",
            "field": "nppv",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "oxygen": {
            "dataset": "med-ogs-bio-rean-d",
            "prefix": "my",
            "field": "o2",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "ph": {
            "dataset": "med-ogs-car-rean-d",
            "prefix": "my",
            "field": "ph",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "dissolved-inorganic-carbon": {
            "dataset": "med-ogs-car-rean-d",
            "prefix": "my",
            "field": "dissic",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "alkalinity": {
            "dataset": "med-ogs-car-rean-d",
            "prefix": "my",
            "field": "talk",
            "slice_mode": "longitude-latitude",
            "has_depth": True,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "surface-partial-pressure-co2": {
            "dataset": "med-ogs-co2-rean-d",
            "prefix": "my",
            "field": "spco2",
            "slice_mode": "longitude-latitude",
            "has_depth": False,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
        "surface-co2-flux": {
            "dataset": "med-ogs-co2-rean-d",
            "prefix": "my",
            "field": "fpco2",
            "slice_mode": "longitude-latitude",
            "has_depth": False,
            "date_limit": datetime.datetime.strptime("01-01-1999", "%d-%m-%Y"),
        },
    }

    def __init__(self, num_retries=10):
        # The following code is adapted from
        # https://help.marine.copernicus.eu/en/articles/4854800-how-to-manipulate-copernicus-marine-data-using-python
        self.num_retries = num_retries
        cas_url = "https://cmems-cas.cls.fr/cas/login"
        username = os.getenv("CMEMS_USERNAME")
        if not username:
            logging.warning("Undefined environment variable CMEMS_USERNAME")
            username = input("Please provide the login for accessing CMEMS : ")
        password = os.getenv("CMEMS_PASSWORD")
        if not password:
            logging.warning("Undefined environment variable CMEMS_PASSWORD")
            password = getpass.getpass(
                "Please provide the password for accessing CMEMS : "
            )
        self.session = setup_session(cas_url, username, password)
        # The following will throw an exception in case of invalid credentials
        if "CASTGC" not in self.session.cookies.get_dict():
            logging.error(
                "Cannot find the CASTGC key in the session object, the credentials are certainly invalid"
            )
            raise ValueError("Invalid credentials")
        self.session.cookies.set("CASTGC", self.session.cookies.get_dict()["CASTGC"])
        logging.info("Connection to cmems successfull")
        self.datastores = {}

    def fetch(self, prefix_dataset, dataset):
        if dataset not in self.datastores:
            logging.info(f"Opening the connection to the dataset {dataset}")
            url = f"https://{prefix_dataset}.cmems-du.eu/thredds/dodsC/{dataset}"
            retries = 0
            while True:
                try:
                    logging.debug(f"Opening {url}")
                    datastream = open_url(
                        url, session=self.session, user_charset="utf-8"
                    )
                    logging.debug(f"Opened with datatype {type(datastream)}")
                    data_store = xr.backends.PydapDataStore(datastream)
                    break
                except:
                    logging.warning(
                        f"Failure, one more retry, {self.num_retries-retries} remaining"
                    )
                    retries += 1
                    if retries >= self.num_retries:
                        raise RuntimeError("Cannot successfully access the dataset")
            self.datastores[dataset] = xr.open_dataset(data_store)
            logging.debug(
                f"The downloaded datastore for {dataset} is : \n {self.datastores[dataset]} "
            )
        return self.datastores[dataset]

    def get_value(
        self,
        date: Union[datetime.datetime, tuple[datetime.datetime, datetime.datetime]],
        long_lat: tuple[float, float],
        depth: Union[float, tuple[float, float]],
        what: str,
        reduction=None,
    ):
        def f_slice(ds, mode, date, long_lat, depth, has_depth):
            if mode == "lon-lat":
                key_lon = "lon"
                key_lat = "lat"
            else:
                key_lon = "longitude"
                key_lat = "latitude"

            slice_coords_definition = {}
            noslice_coords_definition = {
                "method": "nearest",
            }

            if isinstance(long_lat[0], tuple):
                slice_coords_definition[key_lon] = slice(*(long_lat[0]))
            else:
                noslice_coords_definition[key_lon] = long_lat[0]

            if isinstance(long_lat[1], tuple):
                slice_coords_definition[key_lat] = slice(*(long_lat[1]))
            else:
                noslice_coords_definition[key_lat] = long_lat[1]

            if has_depth:
                if isinstance(depth, tuple):
                    slice_coords_definition["depth"] = slice(depth[0], depth[1])
                else:
                    noslice_coords_definition["depth"] = depth

            if isinstance(date, tuple):
                slice_coords_definition["time"] = slice(date[0], date[1])
            else:
                noslice_coords_definition["time"] = date

            # Drop duplicated indices
            # This happens for example with oxygen, nppv, ph, alkalinity, dissic
            ds = ds.drop_duplicates(dim=...)

            # Select the slicing and no slicing from the dataset
            values = ds.sel(**slice_coords_definition)
            values = values.sel(**noslice_coords_definition)

            # Convert the result into a dataframe
            df_values = values.to_dataframe()

            # The noslice coordinates will appear as columns
            # we reset the index to get all the dimensions as columns
            df_values.reset_index(inplace=True)

            if not has_depth:
                # In this case, where a dataset cannot be indexed by depth
                # we simply ignore the depth and fill in the value we got
                # possibly at the surface
                df_values["depth"] = depth

            # before defining our own expected ordering of the dimensions
            df_values.set_index(["time", key_lon, key_lat, "depth"], inplace=True)

            # Ensure we always have longitude and latitude for spatial
            # coordinates
            df_values.index.rename(
                ["time", "longitude", "latitude", "depth"], inplace=True
            )
            if reduction == "mean":
                result = df_values.mean().item()
            else:
                result = df_values

            selected_coordinates = {
                "time": values.coords["time"].to_numpy(),
                key_lon: values.coords[key_lon].to_numpy(),
                key_lat: values.coords[key_lat].to_numpy(),
                "depth": values.coords["depth"].to_numpy()
                if has_depth
                else float("nan"),
            }
            return result, selected_coordinates

        # Get access to the datastore
        if what in CMEMS._feature_params.keys():
            params = CMEMS._feature_params[what]

            # From 1987 to present
            if (isinstance(date, tuple) and date[0] < params["date_limit"]) or (
                not isinstance(date, tuple) and date < params["date_limit"]
            ):
                raise ValueError(f"Cannot get {what} before {params['date_limit']}")
            datastore = self.fetch(params["prefix"], params["dataset"])
            logging.info(f"Slicing for {params['field']}")

            return f_slice(
                datastore[params["field"]],
                params["slice_mode"],
                date,
                long_lat,
                depth,
                params["has_depth"],
            )
        else:
            raise ValueError(
                f"Does not know which dataset to download for the key {what}"
            )
